/** ---------------------------------------------------------
 *  AUDIO bit stream recorder - feat. PortAudio
 *
 * @file silent_record.c
 * @auther rerofumi. <rero2@yuumu.org>
 --------------------------------------------------------- */

/* --------------------------------------
   UPDATE:
   Dec.24.2007 - rerofumi
     first version.
   Dec.12.2007 - rerofumi
     develop start

   ---------------------------------------- */

/*
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* ------------------------ */
/* include                  */
/* ------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include "portaudio.h"


/* ------------------------ */
/* define                   */
/* ------------------------ */

#define SAMPLE_RATE  (11025)
#define FRAMES_PER_BUFFER (1024)
#define NUM_CHANNELS    (1)
/* #define DITHER_FLAG     (paDitherOff) */
#define DITHER_FLAG     (0) /**/

/* --- 8bit sample */
#define PA_SAMPLE_TYPE  paInt16
typedef short SAMPLE;
#define SAMPLE_SILENCE  (0)
#define PRINTF_S_FORMAT "%d"
#define SAMPLE_BYTE_SIZE  2

/* --- sample time (second) */
#define SAMPLESEC  (60*15)
#define BUFFERSIZE ((SAMPLE_RATE*SAMPLE_BYTE_SIZE*SAMPLESEC*NUM_CHANNELS)*3/2)
#define FRAMESIZE (SAMPLE_RATE*SAMPLESEC*3/2)


#define PATHLENGTH  255


/* ------------------------ */
/* struct                   */
/* ------------------------ */

/* --- data store area */
typedef struct
{
  time_t  startTime;
  time_t  endTime;
  int          frameIndex;
  int          maxFrameIndex;
  SAMPLE      *recordedSamples;
} paAudioData;

/* --- double buffer */
typedef struct
{
  /* --- output flag */
  int  iOutput;
  /* --- working buffer */
  int  iPage;
  /* --- sound buffer */
  paAudioData  arrayData[2];
} srBuffer;


/* ------------------------ */
/* global value             */
/* ------------------------ */

/* --- audio data */
srBuffer   sampleBuffer;

/* --- path name */
char  outputpath[PATHLENGTH];


/* ------------------------------------------ */
/*  prototype                                 */
/* ------------------------------------------ */

static int recordCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData );

int writeSampleData(const paAudioData *data, char *filename);

void setRecordTime(paAudioData *data);

/* ------------------------------------------ */
/*  main                                      */
/* ------------------------------------------ */

int main(int argc, char *argv[])
{
  PaStreamParameters  inputParameters;
  PaStream*           stream;
  PaError             err = paNoError;
  const   PaDeviceInfo *deviceInfo;
  struct tm  tmptm;
  int  i, j;
  char filename[PATHLENGTH];
  FILE *pathtest;
  int  numDevice, selectDevice;

  /* --- argment */
  if (argc > 2) {
    printf("Usage: silent_record [output path]\n");
    exit(1);
  }
  else if (argc == 2) {
    strncpy(outputpath, argv[1], PATHLENGTH);
  }
  else {
    /* no arg */
    strncpy(outputpath, ".", PATHLENGTH);
  }

  /* --- path test */
  snprintf(filename, PATHLENGTH-1, "%s/test_tmp", outputpath);
  pathtest = fopen(filename, "w");
  if (pathtest == NULL) {
    fprintf(stderr, "Wrong path name.\n");
    exit(1);
  }
  fclose(pathtest);
  snprintf(filename, PATHLENGTH-1, "rm %s/test_tmp", outputpath);
  system(filename);

  /* --- buffer initialize */
  sampleBuffer.iOutput = 0;
  sampleBuffer.iPage = 0;
  for(i=0; i<2; i++) {
    sampleBuffer.arrayData[i].startTime = 0;
    sampleBuffer.arrayData[i].endTime = 0;
    sampleBuffer.arrayData[i].frameIndex = 0;
    sampleBuffer.arrayData[i].maxFrameIndex = FRAMESIZE;
    sampleBuffer.arrayData[i].recordedSamples = (SAMPLE *) malloc(BUFFERSIZE);
    for(j=0; j<FRAMESIZE; j++) {
      sampleBuffer.arrayData[i].recordedSamples[j] = (SAMPLE)SAMPLE_SILENCE;
    }
    setRecordTime(&sampleBuffer.arrayData[0]);
  }

  /* --- port audio initialize */
  err = Pa_Initialize();
  if( err != paNoError ) goto done;

  /* input device (last device) */
  selectDevice =  Pa_GetDefaultInputDevice();
  numDevice = Pa_GetDeviceCount();
  if (numDevice < 0) {
    fprintf( stderr, "Device not found\n");
  }
  for(i=0; i<numDevice; i++) {
    deviceInfo = Pa_GetDeviceInfo(i);
    if (deviceInfo->maxInputChannels > 0) {
      selectDevice = i;
    }
  }

  /* stereo input */
  inputParameters.device = selectDevice;
  inputParameters.channelCount = NUM_CHANNELS;
  inputParameters.sampleFormat = PA_SAMPLE_TYPE;
  inputParameters.suggestedLatency =
    Pa_GetDeviceInfo( inputParameters.device )->defaultLowInputLatency;
  inputParameters.hostApiSpecificStreamInfo = NULL;

#ifdef DEBUG
  printf("open stream\n");
#endif
  /* --- start record work */
  err = Pa_OpenStream(&stream,
		      &inputParameters,
		      NULL,  /* &outputParameters, */
		      SAMPLE_RATE,
		      FRAMES_PER_BUFFER,
		      paClipOff,   /* we won't output out of range samples so don't bother clipping them */
		      recordCallback,
		      &sampleBuffer );
  if( err != paNoError ) goto done;
  
  err = Pa_StartStream( stream );
  if( err != paNoError ) goto done;
  
  while( ( err = Pa_IsStreamActive( stream ) ) == 1 ) {
    Pa_Sleep(500);
    /* - buffer output */
    if (sampleBuffer.iOutput == 1) {
      sampleBuffer.iOutput = 0;
      i = sampleBuffer.iPage ^ 1;
      localtime_r(&sampleBuffer.arrayData[i].startTime, &tmptm);
      snprintf(filename, PATHLENGTH-1, "%s/%02d%02d%02d%02d.str", 
	       outputpath,
	       tmptm.tm_mon+1,
	       tmptm.tm_mday,
	       tmptm.tm_hour,
	       tmptm.tm_min);
      writeSampleData(&sampleBuffer.arrayData[i], filename);
#ifdef DEBUG
      printf("record file : %s\n", filename);
#endif
    }
  }
  if( err < 0 ) goto done;
  
  err = Pa_CloseStream( stream );
  if( err != paNoError ) goto done;

  fprintf( stderr, "buffer over.\n" );
  
  /* --- end */
 done:
  Pa_Terminate();
  if( sampleBuffer.arrayData[0].recordedSamples != 0) {
    free( sampleBuffer.arrayData[0].recordedSamples );
  }
  if( sampleBuffer.arrayData[1].recordedSamples != 0) {
    free( sampleBuffer.arrayData[1].recordedSamples );
  }
  if( err != paNoError ) {
    fprintf( stderr, "An error occured while using the portaudio stream\n" );
    fprintf( stderr, "Error number: %d\n", err );
    fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
    err = 1;          /* Always return 0 or 1, but no other return codes. */
  }
  return err;
}


/* ------------------------------------------ */
/*  portaudio callback                        */
/* ------------------------------------------ */

static int recordCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData )
{
  srBuffer *buffer = (srBuffer*)userData;
  paAudioData *data = &(buffer->arrayData[buffer->iPage]);
  const SAMPLE *rptr = (const SAMPLE*)inputBuffer;
  SAMPLE *wptr = &data->recordedSamples[data->frameIndex * NUM_CHANNELS];
  long framesToCalc;
  long i;
  int finished;
  unsigned long framesLeft = data->maxFrameIndex - data->frameIndex;
  time_t  now;

  /* Prevent unused variable warnings. */
  (void) outputBuffer;
  (void) timeInfo;
  (void) statusFlags;
  (void) userData;

  if( framesLeft < framesPerBuffer ) {
    framesToCalc = framesLeft;
    finished = paComplete;
  }
  else {
    framesToCalc = framesPerBuffer;
    finished = paContinue;
  }

  if( inputBuffer == NULL ) {
    for( i=0; i<framesToCalc; i++ ) {
      *wptr++ = SAMPLE_SILENCE;  /* left */
      if( NUM_CHANNELS == 2 ) *wptr++ = SAMPLE_SILENCE;  /* right */
    }
  }
  else {
    for( i=0; i<framesToCalc; i++ ) {
      *wptr++ = *rptr++;  /* left */
      if( NUM_CHANNELS == 2 ) *wptr++ = *rptr++;  /* right */
    }
  }

  /* next */
  data->frameIndex += framesToCalc;

  /* --- check time, page swap */
  time(&now);
  if (now >= data->endTime) {
    if (buffer->iPage == 0) {
      buffer->iPage = 1;
    }
    else {
      buffer->iPage = 0;
    }
    buffer->iOutput = 1;
    data = &(buffer->arrayData[buffer->iPage]);
    setRecordTime(data);
    data->frameIndex = 0;
  }

  return (finished);
}


/* ------------------------------------------ */
/*  subroutine                                */
/* ------------------------------------------ */

/**
 * store file
 *
 * @param[in] data sample
 * @param[in] filename filename
 * @return store byte
 * @retval 0 write error
 */
int writeSampleData(const paAudioData *data, char *filename)
{
  int length = data->frameIndex;
  FILE *fp;

  fp = fopen(filename, "wb");
  if (fp == NULL) {
#ifdef DEBUG
    fprintf(stderr, "file open err.\n");
#endif
    return(0);
  }
  else {
    fwrite(data->recordedSamples,
	   SAMPLE_BYTE_SIZE,
	   length,
	   fp);
    fclose(fp);
  }

  return(length);
}


/**
 * time value set
 *
 * @param[in] data  sample data structure
 */
void setRecordTime(paAudioData *data)
{
  time_t  start, end;
  struct tm edittm;
  int  sample_min;

  sample_min = SAMPLESEC / 60;
  time(&start);
  localtime_r(&start, &edittm);
  edittm.tm_sec = 0;
  edittm.tm_min = edittm.tm_min - (edittm.tm_min % sample_min);
  start = mktime(&edittm);
  end = start + SAMPLESEC;
  data->startTime = start;
  data->endTime = end;

#ifdef DEBUG
  char  line[64];
  ctime_r(&start, line);
  printf("startTime = %s", line);
  ctime_r(&end, line);
  printf("endTime = %s", line);
#endif
}

